module registry
  use wrf2gem_subs, only: read_values, mem_error, freq_divisible
  use gempak,       only: write_gempak
  use diagnostics,  only: mix_ratio_to_spec_hum, temp_from_theta_p,  &
                          surface_pres, sea_pres, lifted_index, mean_rh
  use diagnostics,  only: Grav
  implicit none
  save
  private
  public :: init_reg, output_var

  ! The currently programmed WRF variables are:
  ! 1  - HGT: Terrain height (m)
  ! 2  - T2: 2-m temperature (K)
  ! 3  - Q2: 2-m water vapor mixing ratio (kg/kg)
  ! 4  - ZNU: Eta coordinate on mass (half) levels
  ! 5  - PB: Base state pressure (Pa)
  ! 6  - P: Perturbation pressure (Pa)
  ! 7  - T: Perterbation potential temperature (K)
  ! 8  - U: U component of wind (m/s)
  ! 9  - V: V component of wind (m/s)
  ! 10 - W: W component of wind (m/s)
  ! 11 - U10: U at 10 m (m/s)
  ! 12 - V10: V at 10 m (m/s)
  ! 13 - QVAPOR: Water vapor mixing ratio (kg/kg)
  ! 14 - QCLOUD: Cloud water mixing ratio (kg/kg)
  ! 15 - QICE: Ice mixing ratio (kg/kg)
  ! 16 - RAINC: Accumulated cumulus precip (mm)
  ! 17 - RAINNC: Accumulated grid-scale precip (mm)
  ! 18 - PH: Perturbation geopotential (m^2/s^2)
  ! 19 - PHB: Base-state geopotential (m^2/s^2)
  ! 20 - ZNW: Eta coordinate on w (full) levels
  ! 21 - LU_INDEX: Land use category

  ! The currently programmed diagnostics are:
  ! 101 - Surface pressure (hPa)
  ! 102 - Pressure at 2 m (hPa)
  ! 103 - Sea level pressure (hPa)
  ! 104 - Convective precipitation accumulated over one hour (mm)
  ! 105 - Total precipitation accumulated over one hour (mm)
  ! 106 - Total precipitation accumulated over three hours (mm)
  ! 107 - Total precipitation accumulated over six hours (mm)
  ! 108 - Total accumulated precipitation (mm)
  ! 109 - Pressure (hPa)
  ! 110 - Potential temperature (K)
  ! 111 - Unstaggered U component of wind (m/s)
  ! 112 - Unstaggered V component of wind (m/s)
  ! 113 - Cloud water + Ice mixing ratio (kg/kg)
  ! 114 - Unstaggered W component of wind (m/s)
  ! 115 - 2-m temperature including at initial time (K)
  ! 116 - 2-m water vapor mixing ratio including at initial time (kg/kg)
  ! 117 - U at 10 m including at initial time (m/s)
  ! 118 - V at 10 m including at initial time (m/s)
  ! 119 - Lifted Index using lowest 4 levels (K)
  ! 120 - Mean relative humidity (%) in 850-500-hPa layer
  ! 121 - Geopotential height (m) on w (full) levels
  ! 122 - 2-m specific humidity (kg/kg)
  ! 123 - Specific humidity (kg/kg)

  ! Encode GEMPAK vertical coordinate IDs
  integer, parameter :: HAGL = 1279738184, None = 0, Sgma = 4, Pres = 1

  ! Set up enumeration for the available variables
  integer, parameter :: LastWRF = 21, FirstDiag = 101, LastDiag = 123
  integer, parameter :: TerHght   = 1,  &
                        SfcTemp   = 2,  &
                        SfcMixR   = 3,  &
                        EtaMass   = 4,  &
                        PBase     = 5,  &
                        PPert     = 6,  &
                        ThetaPert = 7,  &
                        UStag     = 8,  &
                        VStag     = 9,  &
                        Wstag     = 10,  &
                        SfcU      = 11,  &
                        SfcV      = 12,  &
                        MixR      = 13,  &
                        MixRCloud = 14,  &
                        MixRIce   = 15,  &
                        ConvPre   = 16,  &
                        StabPre   = 17,  &
                        GeopPert  = 18,  &
                        GeopBase  = 19,  &
                        EtaW      = 20,  &
                        LandUse   = 21
  integer, parameter :: SfcP      = 101,  &
                        SfcP2m    = 102,  &
                        SeaLevP   = 103,  &
                        ConvPre1h = 104,  &
                        TotPre1h  = 105,  &
                        TotPre3h  = 106,  &
                        TotPre6h  = 107,  &
                        TotPre    = 108,  &
                        Pressure  = 109,  &
                        Theta     = 110,  &
                        UUnstag   = 111,  &
                        VUnstag   = 112,  &
                        CWtr      = 113,  &
                        WUnstag   = 114,  &
                        SfcTempF0 = 115,  &
                        SfcMixRF0 = 116,  &
                        SfcUF0    = 117,  &
                        SfcVF0    = 118,  &
                        LiftIdx   = 119,  &
                        MeanRH    = 120,  &
                        GeopHghtS = 121,  &
                        SfcSpHum  = 122,  &
                        SpHum     = 123

  ! This structure contains information about the various WRF variables.
  type wrf_var
     integer :: dimType,  &       ! 0-z,t; 1-x,y,t; 2-x,y,z,t; 3-xs,y,z,t;
                                  ! 4-x,ys,z,t; 5-x,y,zs,t
                lev1, lev2,  &    ! GEMPAK level parameters. Relevant only for
                                  ! dimType 1
                vertCordID        ! GEMPAK vertical coordinate ID
     character(len=8) :: wrfName  ! WRF variable name
     character(len=4) :: gemName  ! GEMPAK variable name
     logical          :: const    ! T - Fixed with time; F - Varies with time
  end type wrf_var

  ! This is an analogous structure for diagnosed variables.
  type diag_var
     integer           :: dimType, lev1, lev2, vertCordID, diagID
     character(len=12) :: gemName
  end type diag_var

  type(wrf_var),  dimension(LastWRF)            :: wrfVar
  type(diag_var), dimension(FirstDiag:LastDiag) :: diagVar
  integer                                       :: timesPer12hr

  
contains  ! ===================================================================

  ! init_reg initializes the registry, which contains the vital parameters for
  ! each variable or diagnostic, as well as the number of outputs per 12
  ! hours (freq).
  subroutine init_reg(freq)
    integer, intent(in) :: freq

    wrfVar(TerHght)   = wrf_var(1, 0, -1, None, "HGT     ", "HGHT", .true.)
    wrfVar(SfcTemp)   = wrf_var(1, 2, -1, HAGL, "T2      ", "TMPK", .false.)
    wrfVar(SfcMixR)   = wrf_var(1, 2, -1, HAGL, "Q2      ", "MIXR", .false.)
    wrfVar(EtaMass)   = wrf_var(0, -1, -1, -1, "ZNU     ", "SGMA", .true.)
    wrfVar(PBase)     = wrf_var(2, -1, -1, Sgma, "PB      ", "PBAR", .true.)
    wrfVar(PPert)     = wrf_var(2, -1, -1, Sgma, "P       ", "PP  ", .false.)
    wrfVar(ThetaPert) = wrf_var(2, -1, -1, Sgma, "T       ", "TMPO", .false.)
    wrfVar(UStag)     = wrf_var(3, -1, -1, Sgma, "U       ", "USTG", .false.)
    wrfVar(VStag)     = wrf_var(4, -1, -1, Sgma, "V       ", "VSTG", .false.)
    wrfVar(WStag)     = wrf_var(5, -1, -1, Sgma, "W       ", "WSTG", .false.)
    wrfVar(SfcU)      = wrf_var(1, 10, -1, HAGL, "U10     ", "UREL", .false.)
    wrfVar(SfcV)      = wrf_var(1, 10, -1, HAGL, "V10     ", "VREL", .false.)
    wrfVar(MixR)      = wrf_var(2, -1, -1, Sgma, "QVAPOR  ", "MIXR", .false.)
    wrfVar(MixRCloud) = wrf_var(2, -1, -1, Sgma, "QCLOUD  ", "WC  ", .false.)
    wrfVar(MixRIce)   = wrf_var(2, -1, -1, Sgma, "QICE    ", "WI  ", .false.)
    wrfVar(ConvPre)   = wrf_var(1, 0, -1, None, "RAINC   ", "CTOT", .false.)
    wrfVar(StabPre)   = wrf_var(1, 0, -1, None, "RAINNC  ", "ETOT", .false.)
    wrfVar(GeopPert)  = wrf_var(5, -1, -1, Sgma, "PH      ", "HP  ", .false.)
    wrfVar(GeopBase)  = wrf_var(5, -1, -1, Sgma, "PHB     ", "HBAR", .true.)
    wrfVar(EtaW)      = wrf_var(0, -1, -1, -1, "ZNW     ", "SGMA", .true.)
    wrfVar(LandUse)   = wrf_var(1, 0, -1, None, "LU_INDEX", "LUC ", .true.)

    diagVar(SfcP)      = diag_var(1, 0, -1, None, SfcP, "PRES")
    diagVar(SfcP2m)    = diag_var(1, 2, -1, HAGL, SfcP2m, "PRES")
    diagVar(SeaLevP)   = diag_var(1, 0, -1, None, SeaLevP, "PMSL")
    diagVar(ConvPre1h) = diag_var(1, 0, -1, None, ConvPre1h, "C01M")
    diagVar(TotPre1h)  = diag_var(1, 0, -1, None, TotPre1h, "P01M")
    diagVar(TotPre3h)  = diag_var(1, 0, -1, None, TotPre3h, "P03M")
    diagVar(TotPre6h)  = diag_var(1, 0, -1, None, TotPre6h, "P06M")
    diagVar(TotPre)    = diag_var(1, 0, -1, None, TotPre, "PTOT")
    diagVar(Pressure)  = diag_var(2, -1, -1, Sgma, Pressure, "PRES")
    diagVar(Theta)     = diag_var(2, -1, -1, Sgma, Theta, "THTA")
    diagVar(UUnstag)   = diag_var(2, -1, -1, Sgma, UUnstag, "UREL")
    diagVar(VUnstag)   = diag_var(2, -1, -1, Sgma, VUnstag, "VREL")
    diagVar(CWtr)      = diag_var(2, -1, -1, Sgma, CWtr, "CWTR")
    diagVar(WUnstag)   = diag_var(2, -1, -1, Sgma, WUnstag, "W")
    diagVar(SfcTempF0) = diag_var(1, 2, -1, HAGL, SfcTempF0, "TMPK")
    diagVar(SfcMixRF0) = diag_var(1, 2, -1, HAGL, SfcMixRF0, "MIXR")
    diagVar(SfcUF0)    = diag_var(1, 10, -1, HAGL, SfcUF0, "UREL")
    diagVar(SfcVF0)    = diag_var(1, 10, -1, HAGL, SfcVF0, "VREL")
    diagVar(LiftIdx)   = diag_var(1, 0, -1, None, LiftIdx, "LFT4")
    diagVar(MeanRH)    = diag_var(1, 850, 500, Pres, MeanRH, "RELH")
    diagVar(GeopHghtS) = diag_var(5, -1, -1, Sgma, GeopHghtS, "HGHT")
    diagVar(SfcSpHum)  = diag_var(1, 2, -1, HAGL, SfcSpHum, "SPFH")
    diagVar(SpHum)     = diag_var(2, -1, -1, Sgma, SpHum, "SPFH")
    
    timesPer12hr = freq
  end subroutine init_reg

  ! ===========================================================================

  ! output_var writes a particular WRF variable or diagnostic to the GEMPAK
  ! file associated with gemid.  ncid is the file handle for the WRF history
  ! file from which the WRF data is being read.  The WRF model dimensions are
  ! specified by nx, ny, nz. var is the ID for the requested diagnostic.  times
  ! is an array of output times in GEMPAK form, which corresponds to all of the
  ! times in the history file.
  subroutine output_var(ncid, gemid, nx, ny, nz, var, times)
    integer, intent(in) :: ncid, gemid, nx, ny, nz, var
    character(len=15), dimension(:), intent(in) :: times

    character(len=*), parameter :: fmt1 = "(2x,a,i3,a)"

    write (*,fmt1) "Writing field ", var, " to GEMPAK..."

    ! Call the appropriate subroutine to produce the required output for GEMPAK
    select case (var)
    case (1:LastWRF)
       if (wrfVar(var)%dimType == 1 .or. wrfVar(var)%dimType == 2 .or.  &
            wrfVar(var)%dimType == 5) then
          call simple_out(ncid, gemid, nx, ny, nz, wrfVar(var), times)
       else
          print *, "It does not make sense to output this variable to GEMPAK. &
               &Skipping!"
       end if
    case (FirstDiag:LastDiag)
       if (diagVar(var)%dimType == 1 .or. diagVar(var)%dimType == 2 .or.  &
            diagVar(var)%dimType == 5) then
          call diag_out(ncid, gemid, nx, ny, nz, diagVar(var), times)
       else
          print *, "This diagnostic is not appropriate for output to GEMPAK."
          print *, "It's either staggered in x or y, or not a function of x &
               &and y."
          print *, "Skipping!"
       end if
    case default
       write (*,fmt1) "Variable ", var, " is unknown. Skipping!"
    end select
  end subroutine output_var

  ! ===========================================================================

  ! simple_out reads the variable var in from the WRF history file associated
  ! with ncid and writes it out to the GEMPAK file associated with gemid.
  ! Other inputs to the subroutine are the WRF model dimensions (nx,ny,nz) and
  ! the array of times in GEMPAK form to be output, which corresponds to all of
  ! the times in the history file.
  subroutine simple_out(ncid, gemid, nx, ny, nz, var, times)
    integer,                         intent(in) :: ncid, gemid, nx, ny, nz
    type(wrf_var),                   intent(in) :: var
    character(len=15), dimension(:), intent(in) :: times
     
    real, dimension(:,:,:,:), allocatable :: out4d
    real, dimension(:,:,:),   allocatable :: out3d
    real, dimension(:,:),     allocatable :: elvl
    integer                               :: kMax, eta, nt, status, t, k, lev
    character(len=20)                     :: gdat1
    character(len=12)                     :: parm

    ! Initialization
    select case (var%dimType)
    case (1)  ! x,y,t
       kMax = 0
    case (2)  ! x,y,z,t
       kMax = nz
       eta = EtaMass
    case (5)  ! x,y,zs,t
       kMax = nz + 1
       eta = EtaW
    case default
       ! GEMPAK only understands variables that depend on (unstaggered) x and y
       print *, "Invalid dimType in simple_out!"
       stop
    end select
    
    nt = size(times)
    parm = var%gemName
    
    ! Read variable, and, if necessary, the vertical levels it's at
    if (var%dimType == 1) then
       allocate (out3d(nx,ny,nt), stat=status)
       call mem_error(status, 1, "simple_out")
       call read_values(ncid, var%wrfName, out3d)
    else
       allocate (out4d(nx,ny,kMax,nt), elvl(kMax,nt), stat=status)
       call mem_error(status, 1, "simple_out")
       call read_values(ncid, var%wrfName, out4d)
       call read_values(ncid, wrfVar(eta)%wrfName, elvl)
    end if

    ! Write the variable to GEMPAK
    if (var%const) nt = 1  ! Output constant variables only at initial time
    do t = 1, nt
       gdat1 = times(t)
       if (var%dimType == 1) then
          call write_gempak(gemid, out3d(:,:,t), nx, ny, gdat1, var%lev1,  &
               var%lev2, var%vertCordID, parm)
       else
          do k = 1, kMax
             lev = nint(elvl(k,1) * 10000)
             call write_gempak(gemid, out4d(:,:,k,t), nx, ny, gdat1, lev,  &
                  var%lev2, var%vertCordID, parm)
          end do
       end if
    end do
        
    ! Clean up
    if (allocated(out3d)) then
       deallocate (out3d, stat=status)
    else
       deallocate (out4d, elvl, stat=status)
    end if
    call mem_error(status, 2, "simple_out")
  end subroutine simple_out
  
  ! ===========================================================================

  ! diag_out is a driver routine to produce the diagnostic var based on data
  ! from the WRF history file associated with ncid.  The diagnostic is written
  ! out to the GEMPAK file associated with gemid.  Other inputs to the
  ! subroutine are the WRF model dimensions (nx,ny,nz) and the array of times
  ! in GEMPAK form to be output, which corresponds to all of the times in the
  ! history file.
  subroutine diag_out(ncid, gemid, nx, ny, nz, var, times)
    integer,                         intent(in) :: ncid, gemid, nx, ny, nz
    type(diag_var),                  intent(in) :: var
    character(len=15), dimension(:), intent(in) :: times
    
    real, dimension(:,:,:,:), allocatable :: out4d
    real, dimension(:,:,:),   allocatable :: out3d
    real, dimension(:,:),     allocatable :: elvl
    integer                               :: kMax, eta, nt, status, t, k, lev
    character(len=20)                     :: gdat1
    
    ! Initialization
    select case (var%dimType)
    case (1)  ! x,y,t
       kMax = 0
    case (2)  ! x,y,z,t
       kMax = nz
       eta = EtaMass
    case (5)  ! x,y,zs,t
       kMax = nz + 1
       eta = EtaW
    case default
       print *, "Invalid dimType in diag_out!"
       stop
    end select

    nt = size(times)
  
    if (var%dimType == 1) then
       allocate (out3d(nx,ny,nt), stat=status)
       call mem_error(status, 1, "diag_out")
    else
       allocate (out4d(nx,ny,kMax,nt), elvl(kMax,nt), stat=status)
       call mem_error(status, 1, "diag_out")
       call read_values(ncid, wrfVar(eta)%wrfName, elvl)
    end if
    
    ! Compute appropriate diagnostic
    select case (var%diagID)
    case (SfcP)
       out3d = diag_SfcP(ncid, nx, ny, nz, nt)
    case (SfcP2m)
       out3d = diag_SfcP2m(ncid, nx, ny, nz, nt)
    case (SeaLevP)
       out3d = diag_SeaLevP(ncid, nx, ny, nz, nt)
    case (ConvPre1h)
       if (.not. freq_divisible(timesPer12hr, 12, 1)) then
          print *, "Calculating 1-hr conv. precip makes no sense. Skipping..."
       else
          out3d = diag_ConvPre1h(ncid, nx, ny, nz, nt)
       end if
    case (TotPre1h)
       if (.not. freq_divisible(timesPer12hr, 12, 1)) then
          print *, "Calculating 1-hr total precip makes no sense. Skipping..."
       else
          out3d = diag_TotPre1h(ncid, nx, ny, nz, nt)
       end if
    case (TotPre3h)
       if (.not. freq_divisible(timesPer12hr, 12, 3)) then
          print *, "Calculating 3-hr total precip makes no sense. Skipping..."
       else
          out3d = diag_TotPre3h(ncid, nx, ny, nz, nt)
       end if
    case (TotPre6h)
       if (.not. freq_divisible(timesPer12hr, 12, 6)) then
          print *, "Calculating 6-hr total precip makes no sense. Skipping..."
       else
          out3d = diag_TotPre6h(ncid, nx, ny, nz, nt)     
       end if
    case (TotPre)
       out3d = diag_TotPre(ncid, nx, ny, nz, nt)
    case (Pressure)
       out4d = diag_Pressure(ncid, nx, ny, nz, nt)
    case (Theta)
       out4d = diag_Theta(ncid, nx, ny, nz, nt)
    case (UUnstag)
       out4d = diag_UUnstag(ncid, nx, ny, nz, nt)
    case (VUnstag)
       out4d = diag_VUnstag(ncid, nx, ny, nz, nt)
    case (CWtr)
       out4d = diag_CWtr(ncid, nx, ny, nz, nt)
    case (WUnstag)
       out4d = diag_WUnstag(ncid, nx, ny, nz, nt)
    case (SfcTempF0)
       out3d = diag_SfcTempF0(ncid, nx, ny, nz, nt)
    case (SfcMixRF0)
       out3d = diag_SfcMixRF0(ncid, nx, ny, nz, nt)
    case (SfcUF0)
       out3d = diag_SfcUF0(ncid, nx, ny, nz, nt)
    case (SfcVF0)
       out3d = diag_SfcVF0(ncid, nx, ny, nz, nt)
    case (LiftIdx)
       out3d = diag_LiftIdx(ncid, nx, ny, nz, nt)
    case (MeanRH)
       out3d = diag_MeanRH(ncid, nx, ny, nz, nt)
    case (GeopHghtS)
       out4d = diag_GeopHghtS(ncid, nx, ny, nz, nt)
    case (SfcSpHum)
       out3d = diag_SfcSpHum(ncid, nx, ny, nz, nt)
    case (SpHum)
       out4d = diag_SpHum(ncid, nx, ny, nz, nt)
    case default
       write (*,"(a,i3,a)") "WARNING: Field ", var, " unknown in diag_out!"
    end select

    ! Write the diagnostic out to GEMPAK
    do t = 1, nt
       gdat1 = times(t)
       if (var%dimType == 1) then
          call write_gempak(gemid, out3d(:,:,t), nx, ny, gdat1, var%lev1,  &
               var%lev2, var%vertCordID, var%gemName)
       else
          do k = 1, kMax
             lev = nint(elvl(k,1) * 10000)
             call write_gempak(gemid, out4d(:,:,k,t), nx, ny, gdat1, lev,  &
                  var%lev2, var%vertCordID, var%gemName)
          end do
       end if
    end do

    ! Clean up
    if (allocated(out3d)) then
       deallocate (out3d, stat=status)
    else
       deallocate (out4d, elvl, stat=status)
    end if
    call mem_error(status, 2, "diag_out")
  end subroutine diag_out

  ! ===========================================================================

  ! The next group of functions produce the various diagnostics.  They all take
  ! as input the WRF history file handle (ncid), the WRF domain size 
  ! (nx,ny,nz), and the number of times contained in the history file (nt).

  ! diag_SfcP diagnoses the surface pressure (in hPa).
  function diag_SfcP(ncid, nx, ny, nz, nt) result (ps)
    integer, intent(in)       :: ncid, nx, ny, nz, nt
    real, dimension(nx,ny,nt) :: ps

    character(len=9), parameter :: funName = "diag_SfcP"

    real, dimension(:,:,:,:), allocatable :: p, w, t, ht
    real, dimension(nx,ny,2,nt)           :: ht12
    real, dimension(nx,ny,nt)             :: p1, q1, sfcq, t1
    integer                               :: status
    
    ! Read in pressure, saving only lowest level
    allocate (p(nx,ny,nz,nt), stat=status)
    call mem_error(status, 1, funName//" p")
    p = diag_Pressure(ncid, nx, ny, nz, nt, si=.true.)  ! Do calculations in Pa
    p1 = p(:,:,1,:)
    deallocate (p, stat=status)
    call mem_error(status, 2, funName//" p")

    ! Handle specific humidity
    allocate (w(nx,ny,nz,nt), stat=status)
    call mem_error(status, 1, funName//" w")
    call read_values(ncid, wrfVar(MixR)%wrfName, w)
    q1 = mix_ratio_to_spec_hum(w(:,:,1,:))
    deallocate (w, stat=status)
    call mem_error(status, 2, funName//" w")
    sfcq = mix_ratio_to_spec_hum(diag_SfcMixRF0(ncid, nx, ny, nz, nt))

    ! Handle temperature
    allocate (t(nx,ny,nz,nt), stat=status)
    call mem_error(status, 1, funName//" t")
    t = diag_Theta(ncid, nx, ny, nz, nt)
    t1 = temp_from_theta_p(t(:,:,1,:), p1)
    deallocate (t, stat=status)
    call mem_error(status, 2, funName//" t")

    ! Handle geopotential height
    allocate (ht(nx,ny,nz+1,nt), stat=status)
    call mem_error(status, 1, funName//" ht")
    ht = diag_GeopHghtS(ncid, nx, ny, nz, nt)
    ht12 = ht(:,:,1:2,:)
    deallocate (ht, stat=status)
    call mem_error(status, 2, funName//" ht")

    ps = .01 * surface_pres(ht12, p1, q1, t1, sfcq)  ! Convert to hPa
  end function diag_SfcP

  ! ===========================================================================

  ! diag_SfcP2m diagnoses the pressure (in hPa) 2 meters above the surface.
  ! The approximation is made that rho*g is simply 10 Pa/m.
  function diag_SfcP2m(ncid, nx, ny, nz, nt) result (ps2m)
    integer, intent(in)       :: ncid, nx, ny, nz, nt
    real, dimension(nx,ny,nt) :: ps2m
    
    ps2m = diag_SfcP(ncid, nx, ny, nz, nt) - .2
  end function diag_SfcP2m

  ! ===========================================================================

  ! diag_SeaLevP diagnoses the sea level pressure (in hPa).
  function diag_SeaLevP(ncid, nx, ny, nz, nt) result (slp)
    integer, intent(in)       :: ncid, nx, ny, nz, nt
    real, dimension(nx,ny,nt) :: slp
    
    real, dimension(nx,ny,nz+1,nt) :: ht
    real, dimension(nx,ny,nz,nt)   :: th, p, w
    
    ! Read in WRF data
    ht = diag_GeopHghtS(ncid, nx, ny, nz, nt)
    th = diag_Theta(ncid, nx, ny, nz, nt)
    p = diag_Pressure(ncid, nx, ny, nz, nt, si=.true.)  ! Do calculations in Pa
    call read_values(ncid, wrfVar(MixR)%wrfName, w)  

    ! Compute sea level pressure
    slp = .01 * sea_pres(ht, th, p, mix_ratio_to_spec_hum(w))  ! Convert to hPa
  end function diag_SeaLevP

  ! ===========================================================================
  
  ! diag_ConvPre1h diagnoses 1-hr accumulated convective precipitation (mm).
  function diag_ConvPre1h(ncid, nx, ny, nz, nt) result (convPrecip1h)
    integer, intent(in)       :: ncid, nx, ny, nz, nt
    real, dimension(nx,ny,nt) :: convPrecip1h

    real, dimension(nx,ny,nt) :: accConvPrecip
    integer                   :: k
    
    call read_values(ncid, wrfVar(ConvPre)%wrfName, accConvPrecip)
    convPrecip1h(:,:,1) = 0
    do k = 2, nt
       convPrecip1h(:,:,k) = accConvPrecip(:,:,k) - accConvPrecip(:,:,k-1)
    end do
  end function diag_ConvPre1h

  ! ===========================================================================

  ! diag_TotPre1h diagnoses 1-hr accumulated precipitation (mm).
  function diag_TotPre1h(ncid, nx, ny, nz, nt) result (totPrecip1h)
    integer, intent(in)       :: ncid, nx, ny, nz, nt
    real, dimension(nx,ny,nt) :: totPrecip1h

    real, dimension(nx,ny,nt) :: accTotPrecip
    integer                   :: k
  
    accTotPrecip = diag_TotPre(ncid, nx, ny, nz, nt)
    totPrecip1h(:,:,1) = 0
    do k = 2, nt
       totPrecip1h(:,:,k) = accTotPrecip(:,:,k) - accTotPrecip(:,:,k-1)
    end do
  end function diag_TotPre1h

  ! ===========================================================================

  ! diag_TotPre3h diagnoses 3-hr accumulated precipitation (mm).
  function diag_TotPre3h(ncid, nx, ny, nz, nt) result (totPrecip3h)
    integer, intent(in)       :: ncid, nx, ny, nz, nt
    real, dimension(nx,ny,nt) :: totPrecip3h

    real, dimension(nx,ny,nt) :: accTotPrecip
    integer                   :: threeHrDiff, k

    ! Calculate the time index difference between 3-hr periods
    threeHrDiff = timesPer12hr / 4

    accTotPrecip = diag_TotPre(ncid, nx, ny, nz, nt)
    totPrecip3h(:,:,1:threeHrDiff) = 0
    do k = threeHrDiff+1, nt
       totPrecip3h(:,:,k) =   &
            accTotPrecip(:,:,k) - accTotPrecip(:,:,k-threeHrDiff)
    end do
  end function diag_TotPre3h
  
  ! ===========================================================================

  ! diag_TotPre6h diagnoses 6-hr accumulated precipitation (mm).
  function diag_TotPre6h(ncid, nx, ny, nz, nt) result (totPrecip6h)
    integer, intent(in)       :: ncid, nx, ny, nz, nt
    real, dimension(nx,ny,nt) :: totPrecip6h
    
    real, dimension(nx,ny,nt) :: accTotPrecip
    integer                   :: sixHrDiff, k

    ! Calculate the time index difference between 6-hr periods
    sixHrDiff = timesPer12hr / 2

    accTotPrecip = diag_TotPre(ncid, nx, ny, nz, nt)
    totPrecip6h(:,:,1:sixHrDiff) = 0
    do k = sixHrDiff+1, nt
       totPrecip6h(:,:,k) =  &
            accTotPrecip(:,:,k) - accTotPrecip(:,:,k-sixHrDiff)
    end do
  end function diag_TotPre6h

  ! ===========================================================================

  ! diag_TotPre diagnoses total accumulated precipitation (mm).
  function diag_TotPre(ncid, nx, ny, nz, nt) result (totPrecip)
    integer, intent(in)       :: ncid, nx, ny, nz, nt
    real, dimension(nx,ny,nt) :: totPrecip

    real, dimension(nx,ny,nt) :: accConvPrecip
    real, dimension(nx,ny,nt) :: accStabPrecip
    
    call read_values(ncid, wrfVar(ConvPre)%wrfName, accConvPrecip)
    call read_values(ncid, wrfVar(StabPre)%wrfName, accStabPrecip)
    totPrecip = accConvPrecip + accStabPrecip
  end function diag_TotPre

  ! ===========================================================================

  ! diag_Pressure diagnoses pressure. By default, the output is in hPa, but
  ! if si is true, the output is in Pa.
  function diag_Pressure(ncid, nx, ny, nz, nt, si) result (p)
    integer, intent(in)           :: ncid, nx, ny, nz, nt
    logical, intent(in), optional :: si
    real, dimension(nx,ny,nz,nt)  :: p

    real, dimension(nx,ny,nz,nt) :: pb
    logical                      :: hPa
    
    ! Set appropriate output unit flag
    if (present(si)) then
       hPa = .not. si
    else
       hPa = .true.
    end if

    ! Read in WRF data
    call read_values(ncid, wrfVar(PBase)%wrfName, pb)
    call read_values(ncid, wrfVar(PPert)%wrfName, p)

    ! Calculate pressure
    p = p + pb
    if (hPa) p = .01 * p
  end function diag_Pressure
  
  ! ===========================================================================

  ! diag_Theta diagnoses potential temperature (K).
  function diag_Theta(ncid, nx, ny, nz, nt) result (theta)
    integer, intent(in)          :: ncid, nx, ny, nz, nt
    real, dimension(nx,ny,nz,nt) :: theta
    
    call read_values(ncid, wrfVar(ThetaPert)%wrfName, theta)
    theta = theta + 300  ! WRF stores theta as theta-300
  end function diag_Theta

  ! ===========================================================================

  ! diag_UUnstag diagnoses the wind component (m/s) in the x-coordinate
  ! direction at the mass points (grid box centers).
  function diag_UUnstag(ncid, nx, ny, nz, nt) result (u)
    integer, intent(in)          :: ncid, nx, ny, nz, nt
    real, dimension(nx,ny,nz,nt) :: u

    real, dimension(nx+1,ny,nz,nt) :: us
    integer :: i

    ! Read staggered wind
    call read_values(ncid, wrfVar(UStag)%wrfName, us)

    ! Unstagger it
    do i = 1, nx
       u(i,:,:,:) = .5 * (us(i,:,:,:) + us(i+1,:,:,:))
    end do
  end function diag_UUnstag
  
  ! ===========================================================================

  ! diag_VUnstag diagnoses the wind component (m/s) in the y-coordinate
  ! direction at the mass points (grid box centers).
  function diag_VUnstag(ncid, nx, ny, nz, nt) result (v)
    integer, intent(in)          :: ncid, nx, ny, nz, nt
    real, dimension(nx,ny,nz,nt) :: v

    real, dimension(nx,ny+1,nz,nt) :: vs
    integer :: j

    ! Read staggered wind
    call read_values(ncid, wrfVar(VStag)%wrfName, vs)

    ! Unstagger it
    do j = 1, ny
       v(:,j,:,:) = .5 * (vs(:,j,:,:) + vs(:,j+1,:,:))
    end do
  end function diag_VUnstag
  
  ! ===========================================================================

  ! diag_CWtr diagnoses the cloud water mixing ratio (kg/kg), including ice.
  function diag_CWtr(ncid, nx, ny, nz, nt) result (cw)
    integer, intent(in)          :: ncid, nx, ny, nz, nt
    real, dimension(nx,ny,nz,nt) :: cw

    real, dimension(nx,ny,nz,nt) :: cloud, ice
    
    call read_values(ncid, wrfVar(MixRCloud)%wrfName, cloud)
    call read_values(ncid, wrfVar(MixRIce)%wrfName, ice)
    cw = cloud + ice
  end function diag_CWtr

  ! ===========================================================================
  
  ! diag_WUnstag diagnoses the wind component (m/s) in the vertical direction
  ! at the mass points (grid box centers).
  function diag_WUnstag(ncid, nx, ny, nz, nt) result (w)
    integer, intent(in)          :: ncid, nx, ny, nz, nt
    real, dimension(nx,ny,nz,nt) :: w

    real, dimension(nx,ny,nz+1,nt) :: ws
    integer :: k

    ! Read staggered wind
    call read_values(ncid, wrfVar(WStag)%wrfName, ws)

    ! Unstagger it
    do k = 1, nz
       w(:,:,k,:) = .5 * (ws(:,:,k,:) + ws(:,:,k+1,:))
    end do
  end function diag_WUnstag

  ! ===========================================================================

  ! diag_GeopHghtS diagnoses geopotential height (m) on w (full) levels.
  function diag_GeopHghtS(ncid, nx, ny, nz, nt) result (height)
    integer, intent(in)            :: ncid, nx, ny, nz, nt
    real, dimension(nx,ny,nz+1,nt) :: height
    
    real, dimension(nx,ny,nz+1,nt) :: gpb, gpp
    
    call read_values(ncid, wrfVar(GeopBase)%wrfName, gpb)
    call read_values(ncid, wrfVar(GeopPert)%wrfName, gpp)
    height = (gpb + gpp) / Grav
  end function diag_GeopHghtS

  ! ===========================================================================

  ! diag_SfcTempF0 diagnoses 2-m temperature (K), including the initial time.
  function diag_SfcTempF0(ncid, nx, ny, nz, nt) result (temp)
    integer, intent(in)       :: ncid, nx, ny, nz, nt
    real, dimension(nx,ny,nt) :: temp

    real, dimension(nx,ny,nz,nt) :: potTemp, p
    real, dimension(nx,ny)       :: tk1

    ! For initial time, calculate temperature in lowest grid box.
    potTemp = diag_Theta(ncid, nx, ny, nz, nt)
    p = diag_Pressure(ncid, nx, ny, nz, nt, si = .true.)  ! p in Pa
    tk1 = temp_from_theta_p(potTemp(:,:,1,1), p(:,:,1,1))

    ! Read in WRF 2-m temperature, but WRF doesn't include this variable at
    ! initial time, so assign the temperature in the lowest grid box to it.
    call read_values(ncid, wrfVar(SfcTemp)%wrfName, temp)
    temp(:,:,1) = tk1
  end function diag_SfcTempF0

  ! ===========================================================================

  ! diag_SfcMixRF0 diagnoses 2-m mixing ratio (kg/kg), including the initial
  ! time.
  function diag_SfcMixRF0(ncid, nx, ny, nz, nt) result (w2m)
    integer, intent(in)       :: ncid, nx, ny, nz, nt
    real, dimension(nx,ny,nt) :: w2m

    real, dimension(nx,ny,nz,nt) :: w

    ! Read in WRF 2-m mixing ratio, but WRF doesn't include this variable at
    ! initial time, so assign the mixing ratio in the lowest grid box to it.
    call read_values(ncid, wrfVar(MixR)%wrfName, w)
    call read_values(ncid, wrfVar(SfcMixR)%wrfName, w2m)
    w2m(:,:,1) = w(:,:,1,1)
  end function diag_SfcMixRF0

  ! ===========================================================================

  ! diag_SfcUF0 diagnoses the 2-m wind component (m/s) in the x-coordinate
  ! direction, including the initial time.
  function diag_SfcUF0(ncid, nx, ny, nz, nt) result (u2m)
    integer, intent(in)       :: ncid, nx, ny, nz, nt
    real, dimension(nx,ny,nt) :: u2m
    
    real, dimension(nx,ny,nz,nt) :: u
    
    ! Read in WRF 2-m u-component wind, but WRF doesn't include this variable
    ! at initial time, so assign the u-component wind in the lowest grid box to
    ! it.
    u = diag_UUnstag(ncid, nx, ny, nz, nt)
    call read_values(ncid, wrfVar(SfcU)%wrfName, u2m)
    u2m(:,:,1) = u(:,:,1,1)
  end function diag_SfcUF0

  ! ===========================================================================

  ! diag_SfcVF0 diagnoses the 2-m wind component (m/s) in the y-coordinate
  ! direction, including the initial time.
  function diag_SfcVF0(ncid, nx, ny, nz, nt) result (v2m)
    integer, intent(in)       :: ncid, nx, ny, nz, nt
    real, dimension(nx,ny,nt) :: v2m

    real, dimension(nx,ny,nz,nt) :: v
    
    ! Read in WRF 2-m v-component wind, but WRF doesn't include this variable
    ! at initial time, so assign the v-component wind in the lowest grid box to
    ! it.
    v = diag_VUnstag(ncid, nx, ny, nz, nt)
    call read_values(ncid, wrfVar(SfcV)%wrfName, v2m)
    v2m(:,:,1) = v(:,:,1,1)
  end function diag_SfcVF0

  ! ===========================================================================

  ! diag_LiftIdx diagnoses the Lifted Index (K).
  function diag_LiftIdx(ncid, nx, ny, nz, nt) result (li)
    integer, intent(in)       :: ncid, nx, ny, nz, nt
    real, dimension(nx,ny,nt) :: li

    real, dimension(nx,ny,nz,nt) :: th, p, w
    integer :: i, j, t

    ! Read in the necessary data from WRF output.
    th = diag_Theta(ncid, nx, ny, nz, nt)
    p = diag_Pressure(ncid, nx, ny, nz, nt)  ! p in hPa
    call read_values(ncid, wrfVar(MixR)%wrfName, w)
    
    ! Calculate lifted index, one profile at a time.
    do t = 1, nt
       do j = 1, ny
          do i = 1, nx
             li(i,j,t) = lifted_index(th(i,j,:,t), p(i,j,:,t), w(i,j,:,t))
          end do
       end do
    end do
  end function diag_LiftIdx

  ! ===========================================================================
  
  ! diag_MeanRH diagnoses the mean relative humidity (%) in the 850-500 mb
  ! layer.
  function diag_MeanRH(ncid, nx, ny, nz, nt) result (mrh)
    integer, intent(in)       :: ncid, nx, ny, nz, nt
    real, dimension(nx,ny,nt) :: mrh

    real, dimension(nx,ny,nz,nt) :: th, p, w
    real, dimension(nx,ny,nt)    :: ps
    integer :: i, j, t
    
    ! Read in the necessary data from WRF output.
    th = diag_Theta(ncid, nx, ny, nz, nt)
    p = diag_Pressure(ncid, nx, ny, nz, nt)  ! p in hPa
    call read_values(ncid, wrfVar(MixR)%wrfName, w)
    ps = diag_SfcP(ncid, nx, ny, nz, nt)  ! ps in hPa
    
    ! Calculate mean RH, one profile at a time.
    do t = 1, nt
       do j = 1, ny
          do i = 1, nx
             mrh(i,j,t) = mean_rh(th(i,j,:,t), p(i,j,:,t), w(i,j,:,t),  &
                  ps(i,j,t), 850., 500.)
          end do
       end do
    end do
  end function diag_MeanRH

  ! ===========================================================================

  ! diag_SfcSpHum diagnoses 2-m specific humidity (kg/kg).
  function diag_SfcSpHum(ncid, nx, ny, nz, nt) result (q2m)
    integer, intent(in)       :: ncid, nx, ny, nz, nt
    real, dimension(nx,ny,nt) :: q2m

    q2m = mix_ratio_to_spec_hum(diag_SfcMixRF0(ncid, nx, ny, nz, nt))
  end function diag_SfcSpHum

  ! ===========================================================================
  
  ! diag_SpHum diagnoses specific humidity (kg/kg).
  function diag_SpHum(ncid, nx, ny, nz, nt) result (q)
    integer, intent(in)          :: ncid, nx, ny, nz, nt
    real, dimension(nx,ny,nz,nt) :: q

    real, dimension(nx,ny,nz,nt) :: w

    call read_values(ncid, wrfVar(MixR)%wrfName, w)
    q = mix_ratio_to_spec_hum(w)
  end function diag_SpHum
end module registry
